import {parseToEndpointQueue, ServiceParams} from '../../../common';
import {
	Backend as BaseBackend,
	Server,
	ServerRequest,
	ServerResponse,
	ServerParams,
} from '../../../backend';
import http from 'http';
import { constants } from 'http2';

declare module '../../../backend' {
	interface ServerRequest extends http.IncomingMessage {}

	interface ServerResponse extends http.ServerResponse {}

	interface ImplementationContext {
		res: ServerResponse;
	}
}

class ServerInstance<Backend extends BaseBackend> implements Server<Backend> {
	readonly backend: Backend;
	private serverInternal?: http.Server;

	constructor(params: ServerParams<Backend>) {
		this.backend = params.backend;
	}

	private readonly requestListener = async (req: ServerRequest, res: ServerResponse) => {
		// const endpoints = this.backend.app.endpoints;
		if (typeof req.method === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {});
			res.end();
			return;
		}
		if (typeof req.url === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {});
			res.end();
			return;
		}

		const endpoints = parseToEndpointQueue(req.url, this.backend.app.endpoints);
		const [endpoint, endpointParams] = endpoints.at(-1) ?? [];

		const appOperations = Array.from(this.backend.app.operations)
		const foundAppOperation = appOperations
			.find((op) => op.method === req.method?.toUpperCase());

		if (typeof foundAppOperation === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				'Allow': appOperations.map((op) => op.method).join(',')
			});
			res.end();
			return;
		}

		const endpointOperations = Array.from(endpoint?.operations ?? []);
		if (!endpointOperations.includes(foundAppOperation.name)) {
			res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				'Allow': endpointOperations
					.map((a) => appOperations.find((aa) => aa.name === a)?.method)
					.join(',')
			});
			res.end();
			return;
		}

		const implementation = this.backend.implementations.get(foundAppOperation.name);
		if (typeof implementation === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_NOT_IMPLEMENTED);
			res.end();
			return;
		}

		if (typeof endpoint === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_NOT_IMPLEMENTED);
			res.end();
			return;
		}

		const [, search] = req.url.split('?');

		const responseSpec = await implementation({
			endpoint,
			params: endpointParams ?? {},
			query: typeof search !== 'undefined' ? new URLSearchParams(search) : undefined,
			res,
		});

		if (typeof responseSpec === 'undefined') {
			res.writeHead(constants.HTTP_STATUS_UNPROCESSABLE_ENTITY, {});
			res.end();
			return;
		}

		const finalRes = responseSpec.res ?? res;
		finalRes.statusMessage = responseSpec.statusMessage; // TODO add default status message per status code
		finalRes.writeHead(responseSpec.statusCode, {});
		finalRes.end();
	};

	serve(params: ServiceParams) {
		return new Promise<void>((resolve) => {
			this.serverInternal = new http.Server(this.requestListener);
			this.serverInternal.listen(params.port ?? 80, params.host ?? '0.0.0.0', undefined, resolve);
		});
	}

	close() {
		return new Promise<void>((resolve, reject) => {
			if (typeof this.serverInternal === 'undefined') {
				resolve();
				return;
			}
			this.serverInternal.close((err) => {
				if (err) {
					reject(err);
					return;
				}
				resolve();
			});
		});
	}
}

export const server = <Backend extends BaseBackend>(params: ServerParams<Backend>): Server<Backend> => {
	return new ServerInstance(params);
};
