import {Response} from '../../../backend';
import {ErrorStatusCode, isErrorStatusCode, StatusCode} from '../status-codes';
import http from 'http';

declare module '../../../backend' {
	interface Response {
		statusCode: number;
		statusMessage: string;
		body?: Buffer;
		res?: http.ServerResponse;
	}
}

interface ErrorResponse extends Error, Response {}

interface HttpResponseConstructor<R extends Response> {
	new (...args: any[]): R;
}

interface HttpResponseErrorConstructor<R extends Response> extends HttpResponseConstructor<R> {
	new (message?: string, options?: ErrorOptions): R;
}

interface HttpSuccessResponseConstructor<R extends Response> extends HttpResponseConstructor<R> {
	new (response: Partial<Omit<Response, 'statusCode' | 'res'>>, options?: Pick<Response, 'res'>): R;
}

interface HttpErrorOptions extends ErrorOptions {
	res?: http.ServerResponse;
	body?: Response['body'];
}

export const HttpResponse = <
	T extends StatusCode = StatusCode,
	R extends Response = T extends ErrorStatusCode ? ErrorResponse : Response,
>(statusCode: T): T extends ErrorStatusCode ? HttpResponseErrorConstructor<R> : HttpSuccessResponseConstructor<R> => {
	if (isErrorStatusCode(statusCode)) {
		return class HttpErrorResponse extends Error implements ErrorResponse {
			readonly statusMessage: string;
			readonly statusCode: T;
			readonly res?: http.ServerResponse;
			readonly body?: Buffer;

			constructor(message?: string, options?: HttpErrorOptions) {
				super(message, options);
				this.name = this.statusMessage = message ?? '';
				this.statusCode = statusCode;
				this.cause = options?.cause;
				this.res = options?.res;
				this.body = options?.body;
			}
		} as unknown as HttpResponseErrorConstructor<R>;
	}

	return class HttpSuccessResponse implements Response {
		readonly statusMessage: string;
		readonly statusCode: T;
		readonly body?: Buffer;
		readonly res?: http.ServerResponse;
		constructor(params: Partial<Omit<Response, 'statusCode'>>, options?: Pick<Response, 'res'>) {
			this.statusCode = statusCode;
			this.statusMessage = params.statusMessage ?? '';
			this.body = params.body;
			this.res = options?.res;
		}
	} as unknown as HttpSuccessResponseConstructor<R>;
};
