import {
	App as BaseApp,
	AVAILABLE_EXTENSION_METHODS,
	Endpoint, EndpointQueue,
	GetEndpointParams,
	Operation, serializeEndpointQueue,
	ServiceParams,
} from '../../common';
import {Client, ClientParams, ClientConnection} from '../../client';

declare module '../../client' {
	interface ClientConnection {
		host: string;
		port: number;
		basePath: string;
	}
}

class ClientInstance<App extends BaseApp> implements Client<App> {
	readonly app: App;
	private readonly fetchFn: typeof fetch;
	private connection?: ServiceParams;
	private endpointQueue = [] as EndpointQueue;

	constructor(params: ClientParams<App>) {
		this.app = params.app;
		this.fetchFn = params.fetch ?? fetch;
	}

	async connect(params: ServiceParams): Promise<ClientConnection> {
		const connection = {
			host: params.host ?? '0.0.0.0',
			port: params.port ?? 80,
			basePath: params.basePath ?? '',
		};

		this.connection = connection;

		return connection;
	}

	async disconnect() {
		// noop
	}

	at<TheEndpoint extends Endpoint = Endpoint>(endpoint: TheEndpoint, params?: Record<GetEndpointParams<TheEndpoint>, unknown>) {
		if (Array.isArray(this.endpointQueue)) {
			this.endpointQueue?.push([endpoint, params]);
		}
		return this;
	}

	makeRequest(operation: Operation, query?: URLSearchParams) {
		const baseUrlFragments = [
			this.connection?.host ?? '0.0.0.0'
		];

		const thePort = (this.connection?.port ?? 80);
		if (thePort !== 80) {
			baseUrlFragments.push(thePort.toString());
		}

		const scheme = 'http';
		// todo need a way to decode url back to endpoint queue
		const urlString = serializeEndpointQueue(this.endpointQueue);
		this.endpointQueue = [];

		const url = new URL(
			this.connection?.basePath ? `${this.connection.basePath}${urlString}` : urlString,
			`${scheme}://${baseUrlFragments.join(':')}`
		);
		if (typeof query !== 'undefined') {
			url.search = query.toString();
		}
		const rawEffectiveMethod = (operation.method ?? 'GET').toUpperCase();
		const finalEffectiveMethod = (AVAILABLE_EXTENSION_METHODS as unknown as string[]).includes(rawEffectiveMethod)
			? 'POST' as const
			: rawEffectiveMethod;

		return this.fetchFn(
			url,
			{
				method: finalEffectiveMethod,
			},
		);
	}
}

export const client = <App extends BaseApp>(params: ClientParams<App>): Client<App> => {
	return new ClientInstance(params);
};
